package de.ozgcloud.eingang.semantik.formbased.mantelantrag;

import static de.ozgcloud.eingang.semantik.formbased.mantelantrag.MantelantragZustaendigeStelleMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;

class MantelantragZustaendigeStelleMapperTest {
	private static final String TARGET_OEID = "123456";
	private static final String IDENTIFIER = "gea:test";
	private static final String IDENTIFIER2 = "dfad:test";
	private static final String IDENTIFIER2UPPERCASE = "dfad:Test";
	private static final List<String> IDENTIFIERS = List.of(IDENTIFIER, IDENTIFIER2, "aa:other");

	@Spy
	@InjectMocks
	private MantelantragZustaendigeStelleMapper mapper;

	@BeforeEach
	void mock() {
		doReturn(IDENTIFIERS).when(mapper).getIdentifierList();
		mapper.init();
	}

	@DisplayName("is responsible")
	@Nested
	class TestIsResponsible {

		@DisplayName("should be true if mantelantrag starts with form id")
		@ParameterizedTest
		@ValueSource(strings = { MANTELANTRAG_FORM_ID, MANTELANTRAG_FORM_ID + "_pvog" })
		void shouldBeTrueIfMantelantragStartsWithFormId(String formId) {
			var formData = createWithFormId(formId);

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isTrue();
		}

		@DisplayName("should be false if not mantelantrag form id")
		@Test
		void shouldBeFalseIfNotMantelantragFormId() {
			var formData = createWithFormId("any other id");

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		@DisplayName("should be false if header null")
		@Test
		void shouldBeFalseIfHeaderNull() {
			var formData = FormDataTestFactory.createBuilder()
					.header(null)
					.build();

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		@DisplayName("should be false if form id null")
		@Test
		void shouldBeFalseIfFormIdNull() {
			var formData = FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder()
							.formId(null)
							.build())
					.build();

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		private FormData createWithFormId(String formId) {
			return FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder()
							.formId(formId)
							.build())
					.build();
		}
	}

	@DisplayName("adapt form data")
	@Nested
	class TestAdaptFormData {

		@Mock
		private FormData formData;

		@Mock
		private FormData formData2;

		@DisplayName("should return adapted value")
		@Test
		void shouldReturnAdaptedValue() {
			doReturn(formData2).when(mapper).adaptFormDataWithPossibleException(formData);

			var resultFormData = mapper.parseFormData(formData);

			assertThat(resultFormData).isEqualTo(formData2);
		}

		@DisplayName("should return original value with exception")
		@Test
		void shouldReturnOriginalValueWithException() {
			doThrow(new TechnicalException("some error")).when(mapper).adaptFormDataWithPossibleException(formData);

			var resultFormData = mapper.parseFormData(formData);

			assertThat(resultFormData).isEqualTo(formData);
		}

		@DisplayName("should log warning with exception")
		@Test
		void shouldLogWarningWithException() {
			var exception = new TechnicalException("some error");
			doThrow(exception).when(mapper).adaptFormDataWithPossibleException(formData);

			mapper.parseFormData(formData);

			verify(mapper).logWarningForFailure(exception);
		}
	}

	@DisplayName("adapt form data with possible exception")
	@Nested
	class TestAdaptFormDataWithPossibleException {

		static final String TARGET_OPTIONAL_FIELD_VALUE = "optionalFieldValue";
		static final Map<String, Function<ZustaendigeStelle, String>> FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR = Map.of(
				BEZEICHNUNG_FIELD, ZustaendigeStelle::getBezeichnung,
				EMAIL_FIELD, ZustaendigeStelle::getEmail,
				HAUSANSCHRIFT_STRASSE_FIELD, ZustaendigeStelle::getHausanschriftStrasse,
				HAUSANSCHRIFT_PLZ_FIELD, ZustaendigeStelle::getHausanschriftPlz,
				HAUSANSCHRIFT_ORT_FIELD, ZustaendigeStelle::getHausanschriftOrt,
				TELEFON_FIELD, ZustaendigeStelle::getTelefon
		);
		private FormData formData;
		private Map<String, Object> fieldMap;

		private static Stream<Arguments> fieldNameAndSlotIndex() {
			return Stream.of(0, 1, 2)
					.flatMap(slotIndex -> FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR
							.keySet().stream()
							.map(fieldName -> Arguments.of(slotIndex, fieldName))
					);
		}

		@BeforeEach
		void mock() {
			fieldMap = new HashMap<>(Map.of(
					getOrganisationseinheitIDFieldName(1), " ",
					getOrganisationseinheitIDFieldName(2), "",
					getOrganisationseinheitIDFieldName(10), "1111111"
			));
			formData = FormDataTestFactory.createBuilder()
					.formData(fieldMap)
					.build();
		}

		@DisplayName("should keep form data except zustaendigestelle")
		@Test
		void shouldKeepFormDataExceptZustaendigestelle() {
			doReturn(10).when(mapper).findSlotIndex(fieldMap);

			var resultFormData = mapper.parseFormData(formData);

			assertThat(resultFormData.getId()).isEqualTo(formData.getId());
			assertThat(resultFormData.getHeader()).isEqualTo(formData.getHeader());
			assertThat(resultFormData.getAntragsteller()).isEqualTo(formData.getAntragsteller());
		}

		@DisplayName("should map organisationseinheitId")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldMapOrganisationseinheitId(int slotIndex) {
			mockWithOEID(slotIndex, TARGET_OEID);

			var zustaendigeStelle = getAdaptedZustaendigeStelle();

			assertThat(zustaendigeStelle.getOrganisationseinheitenId()).isEqualTo(TARGET_OEID);
		}

		private String getOrganisationseinheitIDFieldName(int slotIndex) {
			return mapper.getNameForSlotIndex(ORGANISATIONSEINHEIT_ID_FIELD, slotIndex);
		}

		@DisplayName("should throw if organistaionseinheitId is missing")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldThrowIfOrganistationseinheitIdIsMissing(int slotIndex) {
			doReturn(slotIndex).when(mapper).findSlotIndex(fieldMap);

			assertThatThrownBy(this::getAdaptedZustaendigeStelle)
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should throw if organistationseinheitId is empty")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldThrowIfOrganistationseinheitIdIsEmpty(int slotIndex) {
			mockWithOEID(slotIndex, "");

			assertThatThrownBy(this::getAdaptedZustaendigeStelle)
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should map optional field")
		@ParameterizedTest
		@MethodSource("fieldNameAndSlotIndex")
		void shouldMapOptionalField(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);
			fieldMap.put(mapper.getNameForSlotIndex(fieldName, slotIndex), TARGET_OPTIONAL_FIELD_VALUE);

			var zustaendigeStelle = getAdaptedZustaendigeStelle();

			assertThat(getExpectedByFieldName(zustaendigeStelle, fieldName)).isEqualTo(TARGET_OPTIONAL_FIELD_VALUE);
		}

		@DisplayName("should map missing optional fields to empty")
		@ParameterizedTest
		@MethodSource("fieldNameAndSlotIndex")
		void shouldMapMissingOptionalFieldsToEmpty(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);

			var zustaendigeStelle = getAdaptedZustaendigeStelle();

			assertThat(getExpectedByFieldName(zustaendigeStelle, fieldName)).isEmpty();
		}

		@DisplayName("should map empty optional fields to empty")
		@ParameterizedTest
		@MethodSource("fieldNameAndSlotIndex")
		void shouldMapEmptyOptionalFieldsToEmpty(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);
			fieldMap.put(mapper.getNameForSlotIndex(fieldName, slotIndex), "");

			var zustaendigeStelle = getAdaptedZustaendigeStelle();

			assertThat(getExpectedByFieldName(zustaendigeStelle, fieldName)).isEmpty();
		}

		private String getExpectedByFieldName(ZustaendigeStelle zustaendigeStelle, String fieldName) {
			var valueExtractor = FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR.get(fieldName);
			return valueExtractor.apply(zustaendigeStelle);
		}

		private void mockWithSomeOEID(int slotIndex) {
			mockWithOEID(slotIndex, "some");
		}

		private void mockWithOEID(int slotIndex, String targetOeid) {
			fieldMap.put(getOrganisationseinheitIDFieldName(slotIndex), targetOeid);
			doReturn(slotIndex).when(mapper).findSlotIndex(fieldMap);
		}

		private ZustaendigeStelle getAdaptedZustaendigeStelle() {
			var formDataResult = mapper.adaptFormDataWithPossibleException(formData);
			return formDataResult.getZustaendigeStelle();
		}

	}

	@DisplayName("find slot index")
	@Nested
	class TestFindSlotIndex {

		private Map<String, Object> fieldMap;

		@Captor
		private ArgumentCaptor<String> warningCaptor;

		@BeforeEach
		void mock() {
			fieldMap = new HashMap<>(Map.of(
					getZustaendigeStelleName(1), "unknown",
					getZustaendigeStelleName(2), ""
			));
		}

		@DisplayName("should return matching slot")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnMatchingSlot(int slotIndex) {
			fieldMap.put(getZustaendigeStelleName(slotIndex), IDENTIFIER);

			var resultSlotIndex = mapper.findSlotIndex(fieldMap);

			assertThat(resultSlotIndex).isEqualTo(slotIndex);
		}

		@DisplayName("should return matching slot with upper-/lowercase difference in identifiers")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnMatchingSlotWithUppercaseIdentifier(int slotIndex) {
			fieldMap.put(getZustaendigeStelleName(slotIndex), IDENTIFIER2UPPERCASE);

			var resultSlotIndex = mapper.findSlotIndex(fieldMap);

			assertThat(resultSlotIndex).isEqualTo(slotIndex);
		}

		private String getZustaendigeStelleName(int slotIndex) {
			return mapper.getNameForSlotIndex(ZUSTELLUNG_NACHRICHTENBROKER_FIELD, slotIndex);
		}

		@DisplayName("should throw if no slot matches")
		@Test
		void shouldThrowIfNoSlotMatches() {
			assertThatThrownBy(() -> mapper.findSlotIndex(fieldMap))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should log warning if multiple slots match")
		@Test
		void shouldLogWarningIfMultipleSlotsMatch() {
			fieldMap.put(getZustaendigeStelleName(0), IDENTIFIER2);
			fieldMap.put(getZustaendigeStelleName(2), IDENTIFIER);

			mapper.findSlotIndex(fieldMap);

			verify(mapper).logWarningForUnexpected(warningCaptor.capture());
			assertThat(warningCaptor.getValue()).isEqualTo(mapper.getMultipleSlotsMessage(List.of(0, 2)));
		}
	}

	@DisplayName("get name for slot index")
	@Nested
	class TestGetNameForSlotIndex {
		private static Stream<Arguments> nameAndSlotIndexAndExpected() {
			return Stream.of(
					Arguments.of("name", 0, "name"),
					Arguments.of("name", 1, "name2"),
					Arguments.of("name", 2, "name3")
			);
		}

		@DisplayName("should return name")
		@ParameterizedTest
		@MethodSource("nameAndSlotIndexAndExpected")
		void shouldReturnName(String name, int slotIndex, String expected) {
			var resultName = mapper.getNameForSlotIndex(name, slotIndex);

			assertThat(resultName).isEqualTo(expected);
		}
	}
}
